{% extends "page.html" %}

{% block htmltitle -%}
  <title>{{ _("Index") }} - {{ docstitle|striptags|e }}</title>
{% endblock htmltitle -%}

{% macro indexentries(firstname, links) %}
  {%- if links -%}
    <a href="{{ links[0][1] }}">
    {%- if links[0][0] %}<strong>{% endif -%}
    {{ firstname|e }}
    {%- if links[0][0] %}</strong>{% endif -%}
    </a>

    {%- for ismain, link in links[1:] -%}
      , <a href="{{ link }}">{% if ismain %}<strong>{% endif -%}
      [{{ loop.index }}]
      {%- if ismain %}</strong>{% endif -%}
      </a>
    {%- endfor %}
  {%- else %}
    {{ firstname|e }}
  {%- endif %}
{% endmacro %}

{% block content %}
<section class="genindex-section">
  <h1 id="index">{{ _('Index') }}</h1>
  <div class="genindex-jumpbox">
    {%- for key, dummy in genindexentries -%}
    <a href="#{{ key }}"><strong>{{ key }}</strong></a>
    {%- if not loop.last %} | {% endif -%}
    {%- endfor -%}
  </div>
</section>

{%- for key, entries in genindexentries %}
<section id="{{ key }}" class="genindex-section">
  <h2>{{ key }}</h2>
  <table class="max-width indextable genindextable"><tr>
    {%- for column in entries|slice_index(2) if column %}
    <td class="styled-row-col"><ul>
      {%- for entryname, (links, subitems, _) in column %}
        <li>{{ indexentries(entryname, links) }}
        {%- if subitems %}
        <ul>
        {%- for subentryname, subentrylinks in subitems %}
          <li>{{ indexentries(subentryname, subentrylinks) }}</li>
        {%- endfor %}
        </ul>
        {%- endif -%}</li>
      {%- endfor %}
    </ul></td>
    {%- endfor %}
  </tr></table>
</section>
{% endfor %}
{% endblock %}
