/* Prevents two-dimensional scrolling and content loss. */
h1,
code,
li {
    overflow-wrap: break-word;
}
/* Provides padding to push down the "breadcrumb" navigation in nested pages. */
.content {
    padding: 1em 3em;
}
/* Improves spacing around custom sidebar section*/
.sidebar-div {
    margin: var(--sidebar-caption-space-above) 0 0 0;
    padding: var(--sidebar-item-spacing-vertical)
        var(--sidebar-item-spacing-horizontal);
}
/* Custom sidebar heading text. Example: Feedback Section heading. */
.sidebar-heading {
    color: var(--color-sidebar-caption-text);
    font-size: var(--font-size--normal);
    font-weight: 700;
}
/* Improves text used in custom sidebar section. Example: Feedback section content.*/
.sidebar-text {
    color: var(--color-sidebar-caption-text);
    font-size: var(--sidebar-item-font-size);
    line-height: 1.4;
}
/* Removes empty space above the sidebar-tree (under "Feedback" section) */
.sidebar-tree {
    margin-top: 0px;
}
/* Adds padding around AWS Logo in the left sidebar. */
.sidebar-logo {
    padding: 20% 15%;
}
/* Hides a div by default. */
.show-div-sm {
    display: none;
}
/* Hides a div by default. */
.show-div-md {
    display: none;
}
/* Positions items starting from the right. */
.justify-content-right {
    justify-content: right;
}
/* Positions items starting from the left. */
.justify-content-left {
    justify-content: left;
}
/* Hides the sidebar and prevents keyboard users from focusing on elements. */
.hide-sidebar {
    visibility: hidden;
}
/* Hides the icon by default and applies relevant styling. */
.nav-close-icon {
    color: var(--color-foreground-secondary);
    display: flex;
    align-items: center;
    justify-content: center;
    cursor: pointer;
}
/* Aligns theme toggle and menu toggle on a medium viewport */
.content-icon-container {
    align-items: center;
}
/* Adds gap between theme toggle and menu toggle on a small viewport */
.header-right {
    gap: 0.5rem; 
}
/* Safeguard for ensuring menus are always visible on larger screens. */
@media (min-width: 82em) {
    .hide-sidebar {
        visibility: visible;
    }
}
/* We only want to show the close-icon on medium-small screens. */
@media (max-width: 67em) {
    /* Displays a div on a small screen. */
    .show-div-sm {
        display: flex;
    }
}
@media (max-width: 82em) {
    /* Displays a div on a medium screen. */
    .show-div-md {
        display: flex;
    }
}
/* Apply furo styled admonition titles for <h3>. */
h3.admonition-title {
    position: relative;
    margin: 0 -0.5rem 0.5rem;
    padding-left: 2.5rem;
    padding-right: 0.5rem;
    padding-top: 0.4rem;
    padding-bottom: 0.4rem;
    font-weight: 700;
    font-size: 1.5em;
    line-height: 1.25;
    border-radius: unset;
    background-color: var(--color-admonition-title-background);
}
/* Apply furo styled admonition icons before <h3>. */
h3.admonition-title::before {
    content: "";
    position: absolute;
    left: 0.5rem;
    width: 1.5rem;
    height: 1.5rem;
    background-color: var(--color-admonition-title);
    mask-image: var(--icon-admonition-default);
    mask-repeat: no-repeat;
}
/* Hides an element by removing it from the document flow */
.hidden {
    display: none;
}
/* Replaces in-line styling in icons.html */
.icon-opacity{
    opacity: 50%;
}

.max-width {
    width: 100%;
}
/* Ensures columns are aligned properly */
.styled-row-col {
    width: 33%;
    vertical-align: top;
}
/* Applies styling to an error page */
.error-page-style {
    box-sizing: border-box;
    vertical-align: middle;
    text-align: center;
    line-height: 1.5;
    font-family: "Open Sans", Helvetica, Arial, sans-serif;
    font-size: 16px;
    margin: 0;
    padding: 0;
}

.error-page-style h1 {
    font-size: 38px;
    padding: 10px 10px 10px 45px;
    margin: 20px 0 35px -45px;
    font-weight: normal;
}

.error-page-style p {
    padding: 0;
    margin: 0 0 10px;
}
