/*This file replaces the style defined in the "_head_css_variables.html" 
partial from furo==2022.12.7*/

body {
  --color-code-background: #f8f8f8;
  --color-code-foreground: black;
}

/* Dark theme styles */
@media not print {
  body[data-theme="dark"] {
    --color-code-background: #272822;
    --color-code-foreground: #f8f8f2;
  }

  /* For users who prefer dark color scheme */
  @media (prefers-color-scheme: dark) {
    body:not([data-theme="light"]) {
      --color-code-background: #272822;
      --color-code-foreground: #f8f8f2;
    }
  }
}
